#! /usr/bin/env python -u
# coding=utf-8
import glob
import json
import os
import shutil
import sys
import tempfile

from genoml.check_dependencies import check_dependencies
from genoml.parse_arguments import Options
from genoml.steps.data_prune import DataPruneStep
from genoml.steps.model_train import ModelTrainStep
from genoml.steps.model_tune import ModelTuneStep
from genoml.steps.model_validate import ModelValidateStep
from genoml.utils import DescriptionLoader

sys.tracebacklimit = 0


@DescriptionLoader.function_description("genoml")
def cli():
    options = Options("commandline_args.txt")
    dependencies = check_dependencies()

    if options.is_data_prune():
        process = DataPruneStep()
    elif options.is_model_train():
        process = ModelTrainStep()
    elif options.is_model_tune():
        process = ModelTuneStep()
    elif options.is_model_validate():
        process = ModelValidateStep()
    else:
        raise Exception("ISSUE: received unrecognizable option.")

    process.set_environment(options, dependencies)
    process.process()


def error_on_exists(name):
    if os.path.exists(name):
        raise FileExistsError("File exists: '{}'".format(name))


@DescriptionLoader.function_description("genoml")
def train():
    options = Options("train_commandline_args.txt")
    if options.model_file:
        error_on_exists(options.model_file)
    elif options.model_dir:
        error_on_exists(options.model_dir)

    dependencies = check_dependencies()
    tmp_dir = tempfile.mkdtemp()
    options._options['--prune-prefix'] = os.path.join(tmp_dir, "model")
    options._options['--best-model-name'] = "best_model"
    if options.verbose > 0:
        print(tmp_dir)
        print(options._options)

    processes = [DataPruneStep(), ModelTrainStep()]
    if not options.no_tune:
        processes += [ModelTuneStep()]

    for process in processes:
        process.set_environment(options, dependencies)
        process.process()

    model_name = os.path.join(tmp_dir, "model.genoml_model")
    copy_originals(options, tmp_dir)

    if options.model_file:
        shutil.make_archive(model_name, 'bztar', os.path.dirname(options.prefix))
        shutil.move(model_name + ".tar.bz2", options.model_file)
        shutil.rmtree(tmp_dir, True)
    elif options.model_dir:
        shutil.move(tmp_dir, options.model_dir)


@DescriptionLoader.function_description("genoml")
def inference():
    options = Options("inference_commandline_args.txt")
    error_on_exists(options.valid_dir)

    dependencies = check_dependencies()

    os.makedirs(options.valid_dir)
    tmp_dir = tempfile.mkdtemp()
    print(tmp_dir)
    if options.model_file:
        shutil.unpack_archive(options.model_file, tmp_dir, format="bztar")
    elif options.model_dir:
        shutil.rmtree(tmp_dir)
        shutil.copytree(options.model_dir, tmp_dir)

    load_originals(options, tmp_dir)
    options._options['--prune-prefix'] = os.path.join(tmp_dir, "model")
    if options.verbose > 0:
        print(tmp_dir)
        print(options._options)

    for process in [ModelValidateStep()]:
        process.set_environment(options, dependencies)
        process.process()

    for file in glob.glob(tmp_dir + "/model_validation*"):
        shutil.copy(file, options.valid_dir)
    shutil.rmtree(tmp_dir, True)


def copy_originals(options, base_name):
    names = {}
    for group in ["--pheno-file", "--geno-prefix", "--gwas-file"]:
        if group not in options._options or options._options[group] is None or options._options[group] == "":
            continue
        name = group[2:]
        target = os.path.join(base_name, name)
        os.makedirs(target)
        for file in glob.glob(options._options[group] + "*"):
            shutil.copy(file, target)
        names[group] = os.path.basename(options._options[group])
    names["--impute-data"] = options.impute_data
    with open(os.path.join(base_name, "names.json"), "w") as fp:
        json.dump(names, fp)


def load_originals(options, base_name):
    with open(os.path.join(base_name, "names.json")) as fp:
        names = json.load(fp)

    for group, value in names.items():
        if group == "--impute-data":
            options._options["--impute-data"] = value
            continue
        name = group[2:]
        target = os.path.join(base_name, name)
        options._options[group] = os.path.join(target, value)


if __name__ == '__main__':
    cli()



#! /usr/bin/env python -u
# coding=utf-8
import os
import shutil
import subprocess

__author__ = 'Sayed Hadi Hashemi'

test_cases = {
    "prune": {
        "args": "--model-file={model_file} --geno-prefix={geno_prefix} --pheno-file={pheno_file}",
        "validate_args": "--model-file={model_file} --valid-dir={valid_dir} --valid-geno-prefix={valid_geno_prefix} "
                         "--valid-pheno-file={valid_pheno_file} "
    },
    "prsice1gwas": {
        "args": "--model-file={model_file} --geno-prefix={geno_prefix} --pheno-file={pheno_file} --gwas-file={"
                "gwas_file}",
        "validate_args": "--model-file={model_file} --valid-dir={valid_dir} --valid-geno-prefix={valid_geno_prefix} "
                         "--valid-pheno-file={valid_pheno_file} "
    },
    "prsice2gwasCovAddit": {
        "args": "--model-file={model_file} --geno-prefix={geno_prefix} --pheno-file={pheno_file} --gwas-file={"
                "gwas_file} --cov-file={cov_file} --addit-file={add_file}",
        "validate_args": "--model-file={model_file} --valid-dir={valid_dir} --valid-geno-prefix={valid_geno_prefix} "
                         "--valid-pheno-file={valid_pheno_file} --valid-cov-file={valid_cov_file} "
                         "--valid-addit-file={valid_addit_file} "
    },
    "sblup1gwasAdditHerit": {
        "args": "--model-file={model_file} --geno-prefix={geno_prefix} --pheno-file={pheno_file} --gwas-file={"
                "gwas_file} --addit-file={add_file} --herit=0.2 ",
        "validate_args": "--model-file={model_file} --valid-dir={valid_dir} --valid-geno-prefix={valid_geno_prefix} "
                         "--valid-pheno-file={valid_pheno_file} --valid-addit-file={valid_addit_file} "
    }
}

files_to_check = [
    "_confMat.txt",
    "_confMatAtBestThresh_training.txt",
    "_tuned_confMat.txt",
    ".PRS_summary.txt",
    ".training_summary.txt",
    "_validation.validation_summary.txt",
    "_validation_validation_confMat.txt",
    "_validation_confMatAtBestThresh_validation.txt",
]
def test():
    for name, tc in test_cases.items():
        base_dir = f"test_results/{name}/"
        print(f"# {name}\necho ::: {name}")
        # print(f"mkdir -p {base_dir}")
        # print(f"cp -r ../example {base_dir}/")

        opts = {
            "model_file": os.path.join(base_dir, "model.genoml"),
            "valid_dir": os.path.join(base_dir, "valid"),

            "geno_prefix": os.path.join(base_dir, "example/training"),
            "pheno_file": os.path.join(base_dir, "example/training.pheno"),

            "gwas_file": os.path.join(base_dir, "example/example_GWAS.txt"),
            "cov_file": os.path.join(base_dir, "example/training.cov"),
            "add_file": os.path.join(base_dir, "example/training.addit"),

            "valid_geno_prefix": os.path.join(base_dir, "example/validation"),
            "valid_pheno_file": os.path.join(base_dir, "example/validation.pheno"),

            "valid_cov_file": os.path.join(base_dir, "example/validation.cov"),
            "valid_addit_file": os.path.join(base_dir, "example/validation.addit")
        }

        args = tc["args"].format(**opts) + " --n-cores=16"
        validate_args = tc["validate_args"].format(**opts) + " --n-cores=16"

        # print(f"genoml-train {args}")
        # print(f"genoml-inference {validate_args}")
        # print(f"tar -xf {opts['model_file']} -C {opts['valid_dir']}")

        ground_truth_dir = os.path.join("ground_truth", name, "model")
        result_dir = os.path.join(opts["valid_dir"], "model")
        for filename in files_to_check:
            print(f'diff -q {result_dir + filename} {ground_truth_dir + filename}')
        print()


if __name__ == "__main__":
    test()



#! /usr/bin/env python -u
# coding=utf-8
from shutil import copyfile

from genoml.steps import PhenoScale, StepBase
from genoml.utils import DescriptionLoader

__author__ = 'Sayed Hadi Hashemi'


class ModelValidateStep(StepBase):
    """performs validation with existing data"""
    _valid_prefix = None

    def _reduce_validate(self):
        self.execute_command([
            self._dependecies["Plink"],
            "--bfile", self._opt.valid_geno_dir,
            "--extract", f"{self._opt.prune_prefix}.reduced_genos_snpList",
            "--recode", "A",
            "--out", f"{self._valid_prefix}.reduced_genos"
        ], name="Plink")

    def _merge(self):
        self.merge_reduced()

    def _main(self):
        if self._opt.pheno_scale == PhenoScale.DISCRETE:
            script_name = self._opt.VALIDATE_DISC if self._opt.pheno_scale == PhenoScale.DISCRETE \
                else self._opt.VALIDATE_CONT
            self.execute_command([
                self._dependecies["R"],
                script_name,
                self._valid_prefix,
                self._opt.n_cores,
                self._opt.impute_data,
                self._opt.prune_prefix #todo: new best_model
            ], name="VALIDATE_CONT, please make sure you have included .cov and .addit validation files, if used for "
                    "training.")

    @DescriptionLoader.function_description("validation_step")
    def process(self):
        self._valid_prefix = f"{self._opt.prune_prefix}_validation"
        self.model_validate()

    def merge_reduced(self):
        self.execute_command([
            self._dependecies["R"],
            self._opt.MERGE,
            self._opt.valid_geno_dir,
            self._opt.valid_pheno_file,
            self.xna(self._opt.valid_cov_file),
            self.xna(self._opt.valid_addit_file),
            self._valid_prefix
        ], name="R")

    @staticmethod
    def xna(s):
        return s if s is not None else "NA"

    # todo: new
    def model_validate(self):
        """this function performs validation with existing data"""
        # check if GWAS is present (meaning it was also present in training), otherwise the Prune option has been used
        # TODO: find a way to ensure user is providing GWAS for validation, in case it was used during training
        if self._opt.gwas_file is None:
            # we need to specify the forced allele here from the training set genotype file, this pulls the allele to
            # force
            # TODO: refactor to a Python code
            self.cut_column(self._opt.geno_prefix + ".bim",
                            "2,5",
                            self._opt.prune_prefix + ".allelesToForce")

            # plink
            self.execute_command([
                self._dependecies["Plink"],
                "--bfile", self._opt.valid_geno_dir,
                "--extract", self._opt.prune_prefix + '.reduced_genos_snpList',
                "--recode", "A",
                "--recode-allele", self._opt.prune_prefix + '.allelesToForce',
                "--out", self._valid_prefix + '.reduced_genos'
            ], name="model_validate")
        else:  # gwas_file is not None
            # plink
            self.execute_command([
                self._dependecies["Plink"],
                "--bfile", self._opt.valid_geno_dir,
                "--extract", self._opt.prune_prefix + '.reduced_genos_snpList',
                "--recode", "A",
                "--recode-allele", self._opt.prune_prefix + '.variantWeightings',
                "--out", self._valid_prefix + '.reduced_genos'
            ], name="model_validate")

            # copy
            copyfile(self._opt.prune_prefix + ".temp.snpsToPull2", self._valid_prefix + ".temp.snpsToPull2")

            self.execute_command([
                self._dependecies["R"],
                self._opt.SCALE_VAR_DOSES_VALID,
                self._opt.prune_prefix,
                self._opt.gwas_file,
                self._opt.valid_geno_dir,
                self._opt.geno_prefix
            ], name="validate")

        self.merge_reduced()
        self._main()

        self.execute_command([
            self._dependecies["R"],
            self._opt.CHECK_VALIDATION,
            self._valid_prefix,
            self._opt.valid_pheno_file
        ], name="CHECK_VALIDATION")



#! /usr/bin/env python -u
# coding=utf-8
import csv
import glob
import os
import subprocess
import sys

from pandas_plink import read_plink

from genoml.steps import StepBase, PhenoScale
from genoml.utils import DescriptionLoader

__author__ = 'Sayed Hadi Hashemi'


class DataPruneStep(StepBase):
    herit = None
    num_snps = None
    num_samples = None

    @DescriptionLoader.function_description("prune_step")
    def process(self):
        self.check_inputs()
        self.check_geno()
        self.reduce()
        self.merge_reduced()

    def reduce(self):
        # main for prune()
        if self._opt.geno_prefix and self._opt.pheno_file:
            if not (self._opt.cov_file or self._opt.gwas_file or self.herit):
                self.reduce_prune()
            elif self._opt.gwas_file and not self.herit:
                self.reduce_prsice()
                self.scale_variant()
            elif self._opt.gwas_file and self.herit and not self._opt.cov_file:
                self.reduce_sblup()
                self.scale_variant()
            else:
                raise RuntimeError("ISSUE: all files present: geno, pheno, gwas, cov, herit. "
                                   "Need to decide between cov and herit for method.")

    def scale_variant(self):
        self.execute_command([
            self._dependecies["R"],
            self._opt.SCALE_VAR_DOSES_TRAIN,
            self._opt.prefix,
            self._opt.gwas_file,
            self._opt.geno_prefix
        ], name="scaleVariantDoses")

    @DescriptionLoader.function_description("reduce_prune")
    def reduce_prune(self):
        self.execute_command([
            self._dependecies["Plink"],
            "--bfile", self._opt.geno_prefix,
            "--indep-pairwise", "10000",
            "1", "0.1",
            "--out", self._opt.prefix + '.temp'
        ], name="prune")

        self.execute_command([
            self._dependecies["Plink"],
            "--bfile", self._opt.geno_prefix,
            "--extract", self._opt.prefix + '.temp.prune.in',
            "--recode", "A",
            "--out", self._opt.prefix + '.reduced_genos'
        ], name="Plink")

        with open(self._opt.prefix + ".reduced_genos_snpList", "w") as fp:
            subprocess.check_call(["cut", "-f", "1", self._opt.prefix + ".temp.prune.in"], stdout=fp)

    @DescriptionLoader.function_description("reduce_prsice")
    def reduce_prsice(self):
        if self._opt.pheno_scale == PhenoScale.DISCRETE and not self._opt.cov_file:
            print("prsice, disc, cov=na")
            self.execute_command([
                self._dependecies["R"],
                self._opt.PRSICE_R,
                "--binary-target", "T",
                "--prsice", self._dependecies["PRSice"],
                "-n", self._opt.n_cores,
                "--out", self._opt.prefix + ".temp",
                "--pheno-file", self._opt.pheno_file,
                "-t", self._opt.geno_prefix,
                "-b", self._opt.gwas_file,
                "--print-snp",
                "--score", "std",
                "--perm", "10000",
                "--bar-levels",
                "5E-8,4E-8,3E-8,2E-8,1E-8,9E-7,8E-7,7E-7,6E-7,5E-7,4E-7,3E-7,2E-7,1E-7,9E-6,8E-6,7E-6,6E-6,5E-6,4E-6,"
                "3E-6,2E-6,1E-6,9E-5,8E-5,7E-5,6E-5,5E-5,4E-5,3E-5,2E-5,1E-5,9E-4,8E-4,7E-4,6E-4,5E-4,4E-4,3E-4,2E-4,"
                "1E-4,9E-3,8E-3,7E-3,6E-3,5E-3,4E-3,3E-3,2E-3,1E-3,9E-2,8E-2,7E-2,6E-2,5E-2",
                "--no-full",
                "--fastscore",
                "--binary-target", "T",
                "--beta",
                "--snp", "SNP",
                "--A1", "A1",
                "--A2", "A2",
                "--stat", "b",
                "--se", "se",
                "--pvalue", "p"], name="Prune")

        elif self._opt.pheno_scale == PhenoScale.DISCRETE and self._opt.cov_file:
            self.execute_command([
                self._dependecies["R"],
                self._opt.PRSICE_R,
                "--binary-target", "T",
                "--prsice", self._dependecies["PRSice"],
                "-n", self._opt.n_cores,
                "--out", self._opt.prefix + ".temp",
                "--pheno-file", self._opt.pheno_file,
                "--cov-file", self._opt.cov_file,
                "-t", self._opt.geno_prefix,
                "-b", self._opt.gwas_file,
                "--print-snp",
                "--score", "std",
                "--perm", "10000",
                "--bar-levels",
                "5E-8,4E-8,3E-8,2E-8,1E-8,9E-7,8E-7,7E-7,6E-7,5E-7,4E-7,3E-7,2E-7,1E-7,9E-6,8E-6,7E-6,6E-6,5E-6,4E-6,"
                "3E-6,2E-6,1E-6,9E-5,8E-5,7E-5,6E-5,5E-5,4E-5,3E-5,2E-5,1E-5,9E-4,8E-4,7E-4,6E-4,5E-4,4E-4,3E-4,2E-4,"
                "1E-4,9E-3,8E-3,7E-3,6E-3,5E-3,4E-3,3E-3,2E-3,1E-3,9E-2,8E-2,7E-2,6E-2,5E-2",
                "--no-full",
                "--fastscore",
                "--binary-target", "T",
                "--beta",
                "--snp", "SNP",
                "--A1", "A1",
                "--A2", "A2",
                "--stat", "b",
                "--se", "se",
                "--pvalue", "p"
            ], name="Prune")
        elif self._opt.pheno_scale == PhenoScale.CONTINUOUS and not self._opt.cov_file:
            print("prsice, cont, cov=na")
            self.execute_command([
                self._dependecies["R"],
                self._opt.PRSICE_R,
                "--prsice", self._opt.PRSICE,
                "-n", self._opt.n_cores,
                "--out", self._opt.prefix + ".temp",
                "--pheno-file", self._opt.pheno_file,
                "-t", self._opt.geno_prefix,
                "-b", self._opt.gwas_file,
                "--print-snp",
                "--score", "std",
                "--perm", "10000",
                "--bar-levels",
                "5E-8,4E-8,3E-8,2E-8,1E-8,9E-7,8E-7,7E-7,6E-7,5E-7,4E-7,3E-7,2E-7,1E-7,9E-6,8E-6,7E-6,6E-6,5E-6,4E-6,"
                "3E-6,2E-6,1E-6,9E-5,8E-5,7E-5,6E-5,5E-5,4E-5,3E-5,2E-5,1E-5,9E-4,8E-4,7E-4,6E-4,5E-4,4E-4,3E-4,2E-4,"
                "1E-4,9E-3,8E-3,7E-3,6E-3,5E-3,4E-3,3E-3,2E-3,1E-3,9E-2,8E-2,7E-2,6E-2,5E-2",
                "--no-full",
                "--fastscore",
                "--binary-target", "F",
                "--beta",
                "--snp", "SNP",
                "--A1", "A1",
                "--A2", "A2",
                "--stat", "b",
                "--se", "se",
                "--pvalue", "p"
            ], name="Prune")
        elif self._opt.pheno_scale == PhenoScale.CONTINUOUS and self._opt.cov_file:
            print("prsice, cont, cov!=na")
            self.execute_command([
                self._dependecies["R"],
                self._opt.PRSICE_R,
                "--prsice", self._opt.PRSICE,
                "-n", self._opt.n_cores,
                "--out", self._opt.prefix + ".temp",
                "--pheno-file", self._opt.pheno_file,
                "--cov-file", self._opt.cov_file,
                "-t", self._opt.geno_prefix,
                "-b", self._opt.gwas_file,
                "--print-snp",
                "--score", "std",
                "--perm", "10000",
                "--bar-levels",
                "5E-8,4E-8,3E-8,2E-8,1E-8,9E-7,8E-7,7E-7,6E-7,5E-7,4E-7,3E-7,2E-7,1E-7,9E-6,8E-6,7E-6,6E-6,5E-6,4E-6,"
                "3E-6,2E-6,1E-6,9E-5,8E-5,7E-5,6E-5,5E-5,4E-5,3E-5,2E-5,1E-5,9E-4,8E-4,7E-4,6E-4,5E-4,4E-4,3E-4,2E-4,"
                "1E-4,9E-3,8E-3,7E-3,6E-3,5E-3,4E-3,3E-3,2E-3,1E-3,9E-2,8E-2,7E-2,6E-2,5E-2",
                "--no-full",
                "--fastscore",
                "--binary-target", "F",
                "--beta",
                "--snp", "SNP",
                "--A1", "A1",
                "--A2", "A2",
                "--stat", "b",
                "--se", "se",
                "--pvalue", "p"
            ], name="Prune")
        else:
            raise RuntimeError("ISSUE: reduce_prsice not a possible condition present.")


        with DescriptionLoader.context("reduce_prsice2"):
            self.cut_column(self._opt.prefix + '.temp.snp',
                            "2",
                            self._opt.prefix + ".temp.snpsToPull")

            # checking threshold from summary: thresh=$(awk 'NR == 2 {print $3}' $prefix.temp.summary)
            with open(self._opt.prefix + ".temp.summary", 'r') as infile:
                column3 = [cols[2] for cols in csv.reader(infile, delimiter="\t")]
                thresh = column3[1]

            # plink
            self.execute_command([
                self._dependecies["Plink"],
                "--bfile", self._opt.geno_prefix,
                "--extract", self._opt.prefix + '.temp.snpsToPull',
                "--clump", self._opt.gwas_file,
                "--clump-p1", thresh,
                "--clump-p2", thresh,
                "--clump-snp-field", "SNP",
                "--clump-field", "p",
                "--clump-r2", "0.1",
                "--clump-kb", "250",
                "--out", self._opt.prefix + '.tempClumps'
            ], name="Prune")

            # TODO: refarctor to a Python code
            self.cut_column(self._opt.prefix + '.tempClumps.clumped',
                            "2",
                            self._opt.prefix + ".temp.snpsToPull2")

            # plink
            self.execute_command([
                self._dependecies["Plink"],
                "--bfile", self._opt.geno_prefix,
                "--extract", self._opt.prefix + '.temp.snpsToPull2',
                "--recode", "A",
                "--out", self._opt.prefix + '.reduced_genos'
            ], name="Prune")

            # TODO: refarctor to a Python code
            self.cut_column(self._opt.prefix + '.temp.snpsToPull2',
                            "1",
                            self._opt.prefix + ".reduced_genos_snpList")

            # checkPrs: only with PRSice, generating more files after the prune
            self.execute_command([
                self._dependecies["R"],
                self._opt.CHECK_PRS, self._opt.prefix, self._opt.pheno_file
            ], name="prune checkPrs")

    @DescriptionLoader.function_description("reduce_sblup")
    def reduce_sblup(self):
        sblup_lambda = self.num_snps * ((1 / self.herit) - 1)

        self.execute_command([
            self._dependecies["Plink"],
            "--bfile", self._opt.geno_prefix,
            "--pheno", self._opt.pheno_file,
            "--make-bed",
            "--out", self._opt.prefix + '.forSblup'
        ], name="Plink")

        self.execute_command([
            self._dependecies["GCTA"],
            "--bfile", self._opt.prefix + '.forSblup', "--cojo-file", self._opt.gwas_file,
            "--cojo-sblup",
            str(sblup_lambda),
            "--cojo-wind", "10000", "--thread-num", self._opt.n_cores, "--out", self._opt.prefix + '.temp'],
            name="GCTA")

        self.execute_command([
            self._dependecies["R"],
            self._opt.FILTER_SBLUP,
            self._opt.prefix], name="R"
        )

        self.execute_command([
            self._dependecies["Plink"],
            "--bfile", self._opt.prefix + '.forSblup',
            "--extract", self._opt.prefix + '.sblupToPull',
            "--indep-pairwise", "10000", "1", "0.1",
            "--out", self._opt.prefix + '.pruning'],
            name="Plink")

        self.execute_command([
            self._dependecies["Plink"],
            "--bfile", self._opt.prefix + '.forSblup',
            "--extract", self._opt.prefix + '.pruning.prune.in',
            "--recode", "A",
            "--out", self._opt.prefix + '.reduced_genos'
        ], name="Plink")

        with open(self._opt.prefix + ".reduced_genos_snpList", "w") as fp:
            subprocess.check_call(["cut", "-f", "1", self._opt.prefix + '.pruning.prune.in'], stdout=fp)

        self.execute_command([
            self._dependecies["R"],
            self._opt.PREP_SBLUP_SCALE,
            self._opt.prefix,
            self._opt.gwas_file,
            self._opt.geno_prefix
        ], name="prune PREP_SBLUP_SCALE")

        sblup_temp_files = glob.glob(self._opt.prefix + '.forSblup.*')
        for filename in sblup_temp_files:
            os.remove(filename)

    @DescriptionLoader.function_description("prune_check_geno")
    def check_inputs(self):
        for ext in ["bed", "bim", "fam"]:
            filename = f"{self._opt.geno_prefix}.{ext}"
            if not os.path.isfile(filename):
                raise RuntimeError(f"{filename} is not found. Exiting program.")

        if not os.path.isfile(self._opt.pheno_file) or not self._opt.pheno_file.endswith('.pheno'):
            raise RuntimeError("No .pheno file found. Exiting program.")

        if self._opt.cov_file is not None:
            if not os.path.isfile(self._opt.cov_file) or not self._opt.cov_file.endswith('.cov'):
                raise RuntimeError("No .cov file found. Exiting program.")

        if self._opt.gwas_file is not None:
            if not os.path.isfile(self._opt.gwas_file):
                raise RuntimeError("No GWAS file found. Exiting program.")

        if self._opt.herit is not None:
            herit = float(self._opt.herit)
            self.herit = herit
            if not 0 < herit <= 1:
                raise RuntimeError("--herit is not in the range of (0,1].")

        if self._opt.addit_file is not None:
            if not os.path.isfile(self._opt.addit_file) or not self._opt.addit_file.endswith('.addit'):
                raise RuntimeError("No .addit file found. Exiting program.")

        if not os.path.exists(os.path.dirname(self._opt.prefix)):
            os.makedirs(os.path.dirname(self._opt.prefix))

    @DescriptionLoader.function_description("prune_check_inputs")
    def check_geno(self):
        """this function checks genotype input, determining the format."""
        (bim, fam, bed) = read_plink(self._opt.geno_prefix)
        self.num_snps = bim.shape[0]
        self.num_samples = fam.iid.unique().size

    @DescriptionLoader.function_description("merge_reduced")
    def merge_reduced(self):
        self.execute_command([
            self._dependecies["R"],
            self._opt.MERGE,
            self._opt.geno_prefix,
            self._opt.pheno_file,
            self.xna(self._opt.cov_file),
            self.xna(self._opt.addit_file),
            self._opt.prefix
        ], name="R")

    @staticmethod
    def xna(s):
        return s if s is not None else "NA"



#! /usr/bin/env python -u
# coding=utf-8
import subprocess
from enum import Enum

from genoml.utils import ColoredBox, ContextScope, DescriptionLoader

__author__ = 'Sayed Hadi Hashemi'


class PhenoScale(Enum):
    DISCRETE = 0
    CONTINUOUS = 1


class StepBase:
    _opt = {}
    _dependecies = {}

    def set_environment(self, options, dependencies):
        self._opt = options
        self._dependecies = dependencies
        ContextScope.set_verbose(self._opt.verbose > 0)

    def execute_command(self, args, name="", output=None):
        with ColoredBox(ColoredBox.BLUE):
            default_output = None if self._opt.verbose > 2 else subprocess.DEVNULL
            output = default_output if output is None else output
            try:
                subprocess.check_call(
                    args,
                    stdout=output,
                    stderr=default_output,
                )
            except subprocess.CalledProcessError:
                print("Running: ", args)
                raise EnvironmentError(f"{name} fail")

    def cut_column(self, input_file, columns, output_file):
        with open(output_file, "w") as fp:
            self.execute_command([
                "cut",
                "-f", columns,
                input_file,
            ], name="Cut", output=fp)



#! /usr/bin/env python -u
# coding=utf-8

from genoml.steps import StepBase, PhenoScale
from genoml.utils import DescriptionLoader

__author__ = 'Sayed Hadi Hashemi'


class ModelTuneStep(StepBase):
    """Performs secondary tunning for the ML model"""

    @DescriptionLoader.function_description("tune_step")
    def process(self):
        script_file = self._opt.TUNE_DISC if self._opt.pheno_scale == PhenoScale.DISCRETE else self._opt.TUNE_CONT
        self.execute_command([
            self._dependecies["R"],
            script_file,
            self._opt.prefix,
            self._opt.n_cores,
            self._opt.cv_reps,
            self._opt.grid_search,
            self._opt.impute_data,
            self._opt.best_model_name
        ], name="R")



#! /usr/bin/env python -u
# coding=utf-8
import csv
import os
import time

import numpy as np
from docopt import docopt

from genoml import __version__

from genoml.steps import PhenoScale


class Options:
    _options = {}

    def __init__(self, commandline_args_file=None):
        self._options = self.parse_args(commandline_args_file)

    @staticmethod
    def parse_args(commandline_args_file):
        arg_file = os.path.join(os.path.dirname(__file__), "misc", commandline_args_file)
        with open(arg_file, "r") as fp:
            return docopt(fp.read(), version=__version__)

    @property
    def pheno_scale(self):
        """this function checks phenotype input, determining the phenotype format.
        Extract the pheno column and find the number of phenotype classes"""
        # TODO(ffaghri1): Add return hint

        with open(self.pheno_file, 'r') as infile:
            column3 = [cols[2] for cols in csv.reader(infile, delimiter="\t")]
            column3 = column3[1:]  # remove header

        column3 = np.array(column3)
        uniq_pheno = np.unique(column3).size

        if uniq_pheno == 1:
            raise ValueError('check_pheno: phenotype file has only one class.')
        elif uniq_pheno == 2:
            # phenotype file is discrete with two classes
            return PhenoScale.DISCRETE
        else:
            # phenotype file is continuous with uniq_pheno classes
            return PhenoScale.CONTINUOUS

    @property
    def geno_prefix(self):
        return self._options.get('--geno-prefix', None)

    @property
    def pheno_file(self):
        return self._options.get('--pheno-file', None)

    @property
    def cov_file(self):
        return self._options.get('--cov-file', None)

    @property
    def gwas_file(self):
        return self._options.get('--gwas-file', None)

    @property
    def herit(self):
        return self._options.get('--herit', None)

    @property
    def addit_file(self):
        return self._options.get('--addit-file', None)

    @property
    def temp_dir(self):
        return self._options.get('--temp-dir', None)

    @property
    def n_cores(self):
        return self._options.get('--n-cores', None)

    @property
    def train_speed(self):
        return self._options.get('--train-speed', None)

    @property
    def cv_reps(self):
        return self._options.get('--cv-reps', None)

    @property
    def impute_data(self):
        return self._options.get('--impute-data', None)

    @property
    def grid_search(self):
        return self._options.get('--grid-search', None)

    @property
    def prune_prefix(self):
        return self._options.get('--prune-prefix', None)

    @property
    def best_model_name(self):
        return self._options.get('--best-model-name', None)

    @property
    def valid_geno_dir(self):
        return self._options.get('--valid-geno-prefix', None)

    @property
    def valid_pheno_file(self):
        return self._options.get('--valid-pheno-file', None)

    @property
    def valid_cov_file(self):
        return self._options.get('--valid-cov-file', None)

    @property
    def valid_addit_file(self):
        return self._options.get('--valid-addit-file', None)

    @property
    def run_id(self):
        return time.strftime("%Y%m%d-%H%M%S")

    @property
    def prefix(self):
        if '--prune-prefix' not in self._options or self._options.get('--prune-prefix', None) is None:
            return os.path.join(self.temp_dir, self.run_id)
        else:
            return self.prune_prefix

    # TODO: temp for paths to external, fix this by using rpy2 and integration
    @property
    def genoml_env(self):
        return os.environ.copy()

    # @property
    # def genoml_env["PATH"](self):
    #     return "/Users/faraz/Downloads/plink_mac_20181202:" + genoml_env["PATH"]  # plink

    @property
    def other_packages(self):
        return os.path.join(os.path.dirname(__file__), "misc", "R")

    @property
    def PLINK(self):
        return os.path.join(self.other_packages, "plink")

    @property
    def GCTA64(self):
        return os.path.join(self.other_packages, "gcta64_mac")

    @property
    def FILTER_SBLUP(self):
        return os.path.join(self.other_packages, "filterSblup.R")

    @property
    def PRSICE_R(self):
        return os.path.join(self.other_packages, "PRSice.R")

    @property
    def PRSICE_LINUX(self):
        return os.path.join(self.other_packages, "PRSice_linux")

    @property
    def MERGE(self):
        return os.path.join(self.other_packages, "mergeForGenoML.R")

    @property
    def TRAIN_DISC(self):
        return os.path.join(self.other_packages, "trainDisc.R")

    @property
    def TRAIN_CONT(self):
        return os.path.join(self.other_packages, "trainCont.R")

    @property
    def TUNE_DISC(self):
        return os.path.join(self.other_packages, "tuneDisc.R")

    @property
    def TUNE_CONT(self):
        return os.path.join(self.other_packages, "tuneCont.R")

    @property
    def VALIDATE_DISC(self):
        return os.path.join(self.other_packages, "validateDisc.R")

    @property
    def VALIDATE_CONT(self):
        return os.path.join(self.other_packages, "validateCont.R")

    @property
    def SCALE_VAR_DOSES_TRAIN(self):
        return os.path.join(self.other_packages, "scaleVariantDoses_training.R")

    @property
    def SCALE_VAR_DOSES_VALID(self):
        return os.path.join(self.other_packages, "scaleVariantDoses_validation.R")

    @property
    def CHECK_PRS(self):
        return os.path.join(self.other_packages, "checkPrs.R")

    @property
    def CHECK_TRAINING(self):
        return os.path.join(self.other_packages, "checkTraining.R")

    @property
    def CHECK_VALIDATION(self):
        return os.path.join(self.other_packages, "checkValidation.R")

    @property
    def PREP_SBLUP_SCALE(self):
        return os.path.join(self.other_packages, "prepSblupToScale.R")

    def is_data_prune(self):
        return self._options.get('data-prune', None)

    def is_model_train(self):
        return self._options.get('model-train', None)

    def is_model_tune(self):
        return self._options.get('model-tune', None)

    def is_model_validate(self):
        return self._options.get('model-validate', None)

    @property
    def model_file(self):
        return self._options.get('--model-file', None)

    @property
    def model_dir(self):
        return self._options.get('--model-dir', None)

    @property
    def valid_dir(self):
        return self._options.get('--valid-dir', None)

    @property
    def verbose(self):
        return self._options.get('-v', 0)

    @property
    def no_tune(self):
        return self._options.get('--no-tune', False)



#! /usr/bin/env python -u
# coding=utf-8
import json
import os
import traceback

__author__ = 'Sayed Hadi Hashemi'
import textwrap


class ColoredBox:
    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    RESET = 39

    def __init__(self, color=None):
        if color is None:
            color = self.GREEN
        self.__color = color

    def __enter__(self):
        print('\033[{}m'.format(self.__color), end="")

    def __exit__(self, exc_type, exc_val, exc_tb):
        print("\x1b[0m", end="")

    @classmethod
    def wrap(cls, text, color):
        return '\033[{}m'.format(color) + text + "\x1b[0m"


class ContextScope:
    indent = 0
    _verbose = False

    def __init__(self, title, description, error):
        self._title = title
        self._description = description
        self._error = error

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None and exc_val is None and exc_tb is None:
            print(f"{self.get_prefix(ColoredBox.GREEN)}{self._title}: {ColoredBox.wrap('[Done]', ColoredBox.GREEN)}")
            print()
            self.remove_indent()
        else:
            print(f"{self.get_prefix(ColoredBox.RED)}{self._title}: {ColoredBox.wrap('[Failed]', ColoredBox.RED)}")
            print(f"{self.indent_text(self._error)}\n")
            self.remove_indent()
            # print(f"{exc_val} {exc_tb}\n")
            traceback.print_exception(exc_type, exc_val, exc_tb)
            # raise exc_val
            exit(1)

    def __enter__(self):
        self.add_indent()
        print(f"{self.get_prefix(ColoredBox.BLUE)}{self._title}")
        if self._verbose:
            # print(f"{self.indent_text(self._description)}\n")
            print(f"{self._description}\n")

    @classmethod
    def add_indent(cls):
        cls.indent += 1

    @classmethod
    def remove_indent(cls):
        cls.indent -= 1

    @classmethod
    def get_prefix(cls, color=None):
        text = "=" * (cls.indent * 4) + "> "
        if color:
            text = ColoredBox.wrap(text, color)
        return text

    @classmethod
    def indent_text(cls, text):
        WIDTH = 70
        indent = max(0, len(cls.get_prefix()) - 2)
        width = WIDTH - indent
        ret = textwrap.fill(text, width)
        ret = textwrap.indent(ret, " " * indent)
        return ret

    @classmethod
    def set_verbose(cls, verbose):
        cls._verbose = verbose


def function_description(title, description, error):
    def wrap(func):
        def func_wrapper(*args, **kwargs):
            with ContextScope(title, description, error):
                return func(*args, **kwargs)

        return func_wrapper

    return wrap


class DescriptionLoader:
    _descriptions = None

    @classmethod
    def _load(cls):
        description_file = os.path.join(os.path.dirname(__file__), "misc", "descriptions.json")
        with open(description_file) as fp:
            cls._descriptions = json.load(fp)

    @classmethod
    def function_description(cls, key):
        title, description, error = cls.get(key)
        return function_description(title, description, error)

    @classmethod
    def get(cls, key):
        if cls._descriptions is None:
            cls._load()
        title = cls._descriptions[key]["title"]
        description = cls._descriptions[key]["description"]
        error = cls._descriptions[key]["error"]
        return title, description, error

    @classmethod
    def context(cls, key):
        title, description, error = cls.get(key)
        return ContextScope(title, description, error)





from setuptools import setup
import os
from genoml import __version__
with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "README.md"), "r") as fp:
    long_description = fp.read()

with open(os.path.join(os.path.dirname(os.path.realpath(__file__)), "requirements.txt"), "r") as fp:
    requirements = fp.read().split("\n")


setup(
    name='genoml',
    version=__version__,
    packages=['genoml', 'genoml.steps'],
    url='https://github.com/genoml/genoml-core',
    license='Apache-2.0',
    author='Sayed Hadi Hashemi',
    author_email='SayedHadiHashemi@gmail.com',
    maintainer='Faraz Faghri',
    maintainer_email='faraz.faghri@gmail.com',
    description='Machine Learning for Genomic',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    entry_points={
        'console_scripts': [
            'genoml-cli=genoml.main:cli',
            'genoml-train=genoml.main:train',
            'genoml-inference=genoml.main:inference',
        ],
    },
    install_requires=requirements,
    package_data={'genoml': ['misc/*', 'misc/*/*']},
)



#! /usr/bin/env python -u
# coding=utf-8
import logging
import os
import platform
import stat
import subprocess
import zipfile
from io import BytesIO
from pathlib import Path

import requests
import rpy2.robjects.packages as rpackages
from rpy2.robjects.packages import importr

from genoml.utils import DescriptionLoader

__author__ = 'Sayed Hadi Hashemi'


def __get_executable_folder():
    key = "GENOML_DEP_DIR"
    if key in os.environ:
        return os.path.abspath(os.environ.get(key))
    else:
        return os.path.join(str(Path.home()), ".genoml", "misc", "executables")


__executable_folder = __get_executable_folder()


@DescriptionLoader.function_description("check_dependencies_R_Packages")
def check_r_packages():
    for name in __R_PACKAGES:
        try:
            importr(name)
        except Exception:
            try:
                utils = rpackages.importr('utils')
                utils.chooseCRANmirror(ind=1)
                utils.install_packages(name, repos="https://cloud.r-project.org")
            except Exception:
                raise EnvironmentError(f"Missing R Package: {name}")


def check_exec(exec_path, *args, absolute_path=False):
    if not absolute_path:
        binary_path = os.path.join(__executable_folder, exec_path)
    else:
        binary_path = exec_path
    if not os.path.exists(binary_path):
        return False

    _ = subprocess.run([binary_path, *args], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    return True


def install_exec(url, exec_path):
    r = requests.get(url, verify=False, stream=True)
    r.raw.decode_content = True
    buffer = BytesIO()
    buffer.write(r.content)
    with zipfile.ZipFile(buffer, "r") as fp:
        fp.extractall(__executable_folder)

    binary_path = os.path.join(__executable_folder, exec_path)
    os.chmod(binary_path, stat.S_IEXEC)


def check_package(name):
    platform_system = platform.system()

    if name not in __DEPENDENCIES:
        raise EnvironmentError(f"Unknown package: {name}")

    if platform_system not in __DEPENDENCIES[name]:
        raise EnvironmentError(f"Unknown supported OK: {platform_system}")

    entry = __DEPENDENCIES[name][platform_system]

    binary_name = entry["binary"]
    args = entry["version_args"]
    url = entry["url"]

    if check_exec(binary_name, *args):
        logging.debug(f"{name} is found")
        return os.path.join(__executable_folder, binary_name)

    logging.warning(f"Installing {name}")
    install_exec(url, binary_name)
    if not check_exec(binary_name, *args):
        logging.warning(f"Failed to run {name} after installation")
        raise EnvironmentError(f"Can not install {name}")
    else:
        return os.path.join(__executable_folder, binary_name)


@DescriptionLoader.function_description("check_dependencies_R")
def check_R():
    r_path = subprocess.check_output("which Rscript", shell=True).decode().strip()
    if not check_exec(r_path, "--version", absolute_path=True):
        raise EnvironmentError("R is not installed")

    check_r_packages()
    return r_path


@DescriptionLoader.function_description("check_dependencies")
def check_dependencies():
    global __DEPENDENCIES
    ret = {}
    for package, data in __DEPENDENCIES.items():
        if "checker" in data:
            ret[package] = data["checker"]()

    return ret


@DescriptionLoader.function_description("check_dependencies_PRSice")
def check_prsice():
    return check_package("PRSice")


@DescriptionLoader.function_description("check_dependencies_GCTA")
def check_gcta():
    return check_package("GCTA")


@DescriptionLoader.function_description("check_dependencies_Plink")
def check_plink():
    return check_package("Plink")


__DEPENDENCIES = {
    "PRSice": {
        "checker": check_prsice,
        "Darwin": {
            "binary": "PRSice_mac",
            "version_args": ["-v"],
            "url": "https://github.com/choishingwan/PRSice/releases/download/2.1.9/PRSice_mac.zip"
        },
        "Linux": {
            "binary": "PRSice_linux",
            "version_args": ["-v"],
            "url": "https://github.com/choishingwan/PRSice/releases/download/2.1.9/PRSice_linux.zip"
        }
    },
    "GCTA": {
        "checker": check_gcta,
        "Darwin": {
            "binary": "gcta_1.91.7beta_mac/bin/gcta64",
            "version_args": ["-v"],
            "url": "https://cnsgenomics.com/software/gcta/gcta_1.91.7beta_mac.zip"
        },
        "Linux": {
            "binary": "gcta_1.91.7beta/gcta64",
            "version_args": ["-v"],
            "url": "https://cnsgenomics.com/software/gcta/gcta_1.91.7beta.zip"
        }
    },
    "Plink": {
        "checker": check_plink,
        "Darwin": {
            "binary": "plink",
            "version_args": ["--version"],
            "url": "http://s3.amazonaws.com/plink1-assets/plink_mac_20181202.zip"
        },
        "Linux": {
            "binary": "plink",
            "version_args": ["--version"],
            "url": "http://s3.amazonaws.com/plink1-assets/plink_linux_x86_64_20181202.zip"
        }
    },
    "R": {
        "checker": check_R,
    },
}

__R_PACKAGES = [
    "caret", "lattice", "ggplot2", "rBayesianOptimization", "plotROC", "pROC", "doParallel", "randomForest", "xgboost",
    "e1071",
]
